from abc import ABC, abstractmethod


class Tokenizer(ABC):
    @abstractmethod
    def __init__(self, debug_level=20):
        pass

    @abstractmethod
    def tokenize(self, word: str) -> None | list[str]:
        pass



from os import path
import logging

import pinyin
import jieba

logging.getLogger("jieba").setLevel(logging.WARNING)
import jieba.posseg as pseg
import chinese_converter
from wordfreq import word_frequency

from autoanki.Dictionary.Dictionary import Dictionary

PATH_TO_FILE = path.join(path.dirname(__file__), "cedict_ts.u8")

# Use the lookup table here: https://github.com/fxsjy/jieba
NOUN = [
    "n",
    "nt",
    "nrt",
    "ORG",
    "nr",
    "PER",
    "ns",
    "nw",
    "nz",
    "f",
    "s",
    "an",
    "zg",
    "j",
]
TIME = ["t", "TIME", "tg", "g"]
ADPOSITION = ["p"]
PRONOUN = ["r", "rz", "rg"]
QUANTIFIER = ["q", "m"]
VERB = ["v", "vd", "vn", "e", "vg", "h"]
ADVERB = ["ad", "d", "aq", "df"]
OTHER = ["vn", "xc", "x"]
ADJECTIVE = ["a", "b", "ng", "ag"]
CONJUNCTION = ["c"]
PARTICLE = ["y", "u", "uj", "uz", "ul", "uv", "ud", "ug"]
PUNCTUATION = ["w"]
IDIOM = ["i", "l"]
SUFFIX = ["k"]
OTHER = ["z", "o", "x"]


class CEDictionary(Dictionary):
    def __init__(self, debug_level):
        """Chinese to English Dictionary
        Parser for the CC-CEDICT Dictionary:
            https://www.mdbg.net/chinese/dictionary?page=cedict
        This is a file containing 122,839 entries
        """
        self.logger = logging.getLogger("autoanki.cedict")
        self.logger.setLevel(debug_level)
        self.logger.info("Loading Chinese Dictionary (CEDictionary)")
        self.dict = self._parse_file()
        self.logger.debug("Done!")

        pass

    def _parse_file(self) -> dict[str, dict]:
        """Parse dictionary file on load
        `return` A dictionary of the information, the key being the Chinese word
        """
        self.logger.debug("Parsing CE-DICT file...")
        if not path.isfile(PATH_TO_FILE):
            self.logger.critical("Could not open dictionary file")

        # TODO Some entries use a dot in the midde
        #   e.g. 哈利·波特
        #   this does not currently match
        definitions = {}
        last_word = "~"
        current_word = ""

        with open(PATH_TO_FILE, "r") as dict_file:
            for line in dict_file:
                if line[0] == "#":
                    continue
                parts = line.split(" ")
                trad_word = parts[0]
                current_word = parts[1]

                definition = line[line.find("]") + 2 :]

                if current_word != last_word:
                    # self.logger.debug(f"Creating entry: {current_word}")
                    definitions[current_word] = {
                        "trad_word": "",
                        "pinyin_numbers": "",
                        "definition": "",
                        "pinyin": "",
                        "frequency": "",
                    }

                # These only need to be done the first time the word is seen
                if last_word != current_word:
                    definitions[current_word]["trad_word"] = trad_word
                    definitions[current_word]["pinyin"] = pinyin.get(current_word)
                    definitions[current_word]["pinyin_numbers"] = pinyin.get(
                        current_word, format="numerical"
                    )
                    definitions[current_word]["frequency"] = str(
                        100 * word_frequency(current_word, "zh")
                    )

                definitions[current_word]["definition"] += "<br>" + definition
                last_word = current_word

        for key, _ in definitions.items():
            definitions[key]["definition"].strip("\n")
            definitions[key]["definition"] = definitions[key]["definition"].lstrip(
                "<br>"
            )
            definitions[key]["definition"] = definitions[key]["definition"].lstrip("/")

            definitions[key]["definition"] = definitions[key]["definition"].rstrip("\n")
            definitions[key]["definition"] = definitions[key]["definition"].rstrip("/")
            # self.logger.info('[' + definitions[key]['definition'] + ']')
        return definitions

    def find_word(self, word: str) -> None | list[str]:
        """
        Find a word in the dictionary. This can be simplified, or traditional
        `return` Paramaters that get passed to the database
        """
        # Convert the word to simplified if needed
        word = chinese_converter.to_simplified(word)
        # self.logger.info(f"[{word}] [{word in self.dict}]")
        if not word in self.dict:
            return None

        """
        traditional_script = params[0]\n
        word_type = params[1]\n
        pinyin = params[2]\n
        pinyin_numbers = params[3]\n
        sub_components = params[4]\n
        hsk_level = params[5]\n
        top_level = params[6]\n
        definition = params[7]\n
        frequency = params[8]
        word = params[9]
        """
        params = ["", "", "", "", "", "", "", "", "", ""]

        slice = next(pseg.cut(word))
        part_of_speech = self.get_part_of_speech(word, slice.flag)
        if part_of_speech != None:
            params[1] = part_of_speech

        params[0] = self.dict[word]["trad_word"]
        params[2] = self.dict[word]["pinyin"]
        params[3] = self.dict[word]["pinyin_numbers"]
        params[7] = self.dict[word]["definition"]
        # self.logger.debug(f"Word: {word}")
        params[8] = self.dict[word]["frequency"]
        params[9] = word
        return params

    def size(self):
        with open(PATH_TO_FILE, "r") as f:
            return len(f.readlines())

    def get_part_of_speech(self, word: str, code: str) -> str:
        if code in NOUN:
            return "noun"
        if code in TIME:
            return "time"
        if code in ADPOSITION:
            return "adposition"
        if code in PRONOUN:
            return "pronoun"
        if code in QUANTIFIER:
            return "quantifier"
        if code in VERB:
            return "verb"
        if code in ADVERB:
            return "adverb"
        if code in OTHER:
            return "other"
        if code in ADJECTIVE:
            return "adjective"
        if code in CONJUNCTION:
            return "conjunction"
        if code in PARTICLE:
            return "particle"
        if code in PUNCTUATION:
            return "punctuation"
        if code in IDIOM:
            return "idiom"
        if code in PARTICLE:
            return "particle"
        if code in OTHER:
            return "other"
        if code in SUFFIX:
            return "suffix"

        self.logger.warning(f"Part of speech code not covered: {word}:{code}")
        return "other"



import os
import sqlite3
import logging

from pathlib import Path

from autoanki.Tokenizer.ChineseTokenizer import ChineseTokenizer
from autoanki.DatabaseManager.DatabaseManager import DatabaseManager


class ChineseDatabaseManager(DatabaseManager):
    def __init__(self, database_path, debug_level, dictionary=None):
        """Interfaces with an SQLLite database containing information on words, definitions, and more"""
        self.logger = logging.getLogger("autoanki.dbmngr")
        self.logger.setLevel(debug_level)

        # Init database
        if not os.path.exists(database_path):
            self.logger.warning("The database [", database_path, "] does not exist.")
            raise Exception("Cannot create DatabaseManager with invalid database path.")
        self.database_path = database_path

        # Text segmenter
        self.tokenizer = ChineseTokenizer(debug_level, dictionary=dictionary)

        # Sql connection
        path = os.path.join(os.getcwd(), self.database_path)
        self.connection = sqlite3.connect(path)
        self.cursor = self.connection.cursor()

        self.books = []

    @staticmethod
    def create_database(database_path: str):
        """Creates autoanki database file, including all tables needed
        Args:
            `database_path`: The path to the database to create
        """
        logger = logging.getLogger("autoanki.dbmngr")
        logger.info("Creating database [" + database_path + "]")
        path = os.path.join(os.path.dirname(__file__), "databases_init.sql")
        try:
            with open(path, "r") as sql_file:
                sql_script = sql_file.read()
            connection = sqlite3.connect(database_path)
            cursor = connection.cursor()
            cursor.executescript(sql_script)
            connection.commit()
            return True
        except FileNotFoundError:
            logger.warning("Could not create database: Missing SQL files")
            return False

    def _create_book_table(self, book_name, table_name) -> bool:
        """Creates a new entry in the book_list table
        Args:
            `table_name`: The name of the new entry to create
        Return:
            False if error
        """
        self.cursor.execute("SELECT table_name FROM book_list")
        book_list = self.cursor.fetchall()

        if (table_name,) in book_list:
            self.logger.warning("The book is already in database. Not adding")
            return False

        # self.logger.debug("Inserting")
        self.cursor.execute(
            f'INSERT INTO book_list VALUES("{book_name}","{table_name}",\'cn\')'
        )
        self.connection.commit()
        fname = "book_table.sql"
        this_file = os.path.abspath(__file__)
        this_dir = os.path.dirname(this_file)
        wanted_file = os.path.join(this_dir, fname)
        fd = open(wanted_file, "r")
        book_table_file = fd.read()
        book_table_file = book_table_file.replace("BOOK_NAME", table_name)
        fd.close()
        # Create the new table
        self.cursor.execute(book_table_file)
        self.connection.commit()
        return True

    def add_contents_to_database(self, contents: str, table_name: str):
        """Adds every word in a file to both the dictionary table and the book's table
        Args:
            `filepath`: The path to the file
            `table_name`: The name of the table to add the words to.
                This should be the same for every wile in a given book
        """
        self.logger.info(f"Adding contents to database...")
        word_appearances = {}
        lines = contents.splitlines()
        for line in lines:
            if not line:
                continue
            words = self.tokenizer.tokenize(line)
            if not words:
                continue
            for word in words:
                if word_appearances.get(word) == None:
                    word_appearances[word] = 1
                else:
                    word_appearances[word] += 1

        # Add the words to the dictionary if they are not already there
        self.cursor.execute(f"SELECT word FROM dictionary")
        self.connection.commit()
        dictionary_words = self.cursor.fetchall()
        self.logger.info(
            f"{len(word_appearances.items())} words in file. {len(dictionary_words)} in dictionary."
        )

        for word, appearances in word_appearances.items():
            if (word,) not in dictionary_words:
                self.insert_word(word)

        # Make a dictionary of word ids from dictionary
        self.cursor.execute(f"SELECT word_id, word FROM dictionary")
        result = self.cursor.fetchall()
        word_id_dict = {}
        for line in result:
            word_id_dict[line[1]] = line[0]

        self.cursor.execute(
            f"SELECT dictionary_word_id, number_of_appearances FROM {table_name}"
        )
        self.connection.commit()
        book_table_response = self.cursor.fetchall()
        book_table_appearances = {}
        # Get the number of appearances in the book table
        for i in book_table_response:
            book_table_appearances[i[0]] = i[1]
        # pprint.pprint(book_table_appearances)

        # Add all words to the book_table
        for word, appearances in word_appearances.items():

            dictionary_word_id = word_id_dict[word]

            # If the word is already in the dictionary, add the number of appearances to it
            if dictionary_word_id in book_table_appearances:
                file_appearances = word_appearances[word]
                db_appearances = book_table_appearances[dictionary_word_id]
                # print("File app:", file_appearances)
                # print("Book app:", book_table_appearances[dictionary_word_id])
                sum = file_appearances + db_appearances
                self.cursor.execute(
                    f"UPDATE {table_name} SET number_of_appearances = ? "
                    f"WHERE dictionary_word_id = ?",
                    [sum, dictionary_word_id],
                )
                self.connection.commit()

            else:
                self.cursor.execute(
                    f"INSERT INTO {table_name} (dictionary_word_id, number_of_appearances) "
                    f"VALUES (?,?)",
                    [dictionary_word_id, word_appearances[word]],
                )
                self.connection.commit()

        # self.logger.debug("Done adding file to database")

    def insert_word(self, word):
        # TODO Doing this breaks the `number_of_appearances`. This is a temporary fix
        # TODO This is really ineficcient
        #   Either contain an internal dict of words in the dictionary,
        #   or wrap this with a try catch
        self.cursor.execute("SELECT word FROM dictionary WHERE word = ?", [word])
        all_rows = self.cursor.fetchall()
        if len(all_rows) == 0:
            # self.logger.info("  Inserting")
            self.cursor.execute(f"INSERT INTO dictionary (word) VALUES (?)", [word])
            self.connection.commit()

    def remove_word(self, word: str):
        # TODO: Make more stringent constraints on removing words. This should be extremely rare
        if "*" in word:
            self.logger.info(f"Not executing: [{word}]. Star in command")
            return
        self.cursor.execute(f"DELETE FROM dictionary WHERE word=(?)", [word])
        self.connection.commit()

    def add_book_from_string(self, contents: str, book_name: str) -> bool:
        return self._add_book(contents, book_name)

    def add_book_from_file(self, filepath: str, book_name: str) -> bool:
        if not os.path.isfile(filepath):
            self.logger.info(f"File does not exist: [{filepath}]")
            return False

        with open(filepath, "r") as file:
            contents = file.read()
            return self._add_book(contents, book_name)

    # def add_book_from_folder(directory):
    # TODO This

    # Legacy code from _add_book()
    # self.logger.debug("Bookpath: " + filepath)
    # if os.path.isdir(filepath):
    #     self.logger.debug("Directory found:")
    #     result = [y for x in os.walk(filepath + '/' + CLEANED_FILES_DIRECTORY) for y in glob(os.path.join(x[0], '*.txt'))]
    #     for path in result:
    #         self.logger.debug(path)
    #         self.add_file_to_database(path, book_tablename)

    def _add_book(self, contents: str, book_name: str) -> bool:
        """Adds a file to the autoanki database.
        This involves the following steps:\n
        1 - Add book to the `book_list` table
        2 - Add all the files in "bookpath" to the definitions table and book table
        Args:
            `bookpath`: The filepath to the book. This is file, or a directory of files
            `book_name: The name of the book. This will show up in the Anki deck
        """
        self.logger.info("Adding book...")
        # Get a 'table name' clean version of the book name
        book_tablename = self.convert_to_tablename(book_name)
        # Add the name of the book to the book_list table
        success = self._create_book_table(book_name, book_tablename)
        if not success:
            self.logger.error(f"Failed to create book table: [{book_tablename}]")
            return False

        # Add all the words in the book to the `definitions` table
        self.add_contents_to_database(contents, book_tablename)

        self.logger.info("Done adding book.")
        return True

    def print_info(self):
        """Print basic information about the database"""
        self.cursor.execute("SELECT word FROM dictionary")
        all_rows = self.cursor.fetchall()
        self.cursor.execute("SELECT word FROM dictionary WHERE definition IS NULL")
        unfinished_rows = self.cursor.fetchall()
        format_string_int = "{:<30} | {:>8}"
        format_string_dec = "{:<30} | {:>8.2}"
        print("------------------------")
        print(self.database_path)
        print("------------------------")
        print(format_string_int.format("Number of books:", len(self.books)))
        print(
            format_string_dec.format(
                "Database size (MB):",
                Path(self.database_path).stat().st_size / (1024 * 1024),
            )
        )
        print("Dictionary Table:")
        print(format_string_int.format("Number of rows:", len(all_rows)))
        print(
            format_string_int.format("Number of unfinished rows:", len(unfinished_rows))
        )

    def update_definition(self, params: list):
        f"""Complete a definition for one word in the dictionary table\n
        traditional_script = params[0]\n
        word_type = params[1]\n
        pinyin = params[2]\n
        pinyin_numbers = params[3]\n
        sub_components = params[4]\n
        hsk_level = params[5]\n
        top_level = params[6]\n
        definition = params[7]\n
        frequency = params[8]
        word = params[9]
        :param params: A list of params for the database:
        :return:
        """
        # TODO Some sanatization here might be a good idea
        #   Make sure junk data can't crash this function
        self.cursor.execute(
            "UPDATE dictionary "
            "SET word_traditional = ?, "
            "word_type = ?,"
            "pinyin = ?, "
            "pinyin_numbers = ?,"
            "sub_components = ?,"
            "hsk_level = ?,"
            "top_level = ?,"
            "definition = ?,"
            "frequency = ?"
            "WHERE word = ?",
            params,
        )
        self.connection.commit()

    def get_all_completed_definitions(self):

        self.cursor.execute("SELECT * FROM dictionary WHERE definition IS NOT NULL")
        raw_words = self.cursor.fetchall()

        # pprint.pp(raw_words)
        words = []
        for row in raw_words:
            word = {
                "word_id": row[0],
                "word": row[1],
                "word_traditional": row[2],
                "word_type": row[3],
                "pinyin": row[4],
                "pinyin_numbers": row[5],
                "number_of_strokes": row[6],
                "sub_components": row[7],
                "frequency": row[8],
                "hsk_level": row[9],
                "top_level": row[10],
                "audio_path": row[11],
                "image_path": row[12],
                "definition": row[13],
            }
            # print(word["word"])
            # pprint.pp(word)
            # words.append(word)
            words.append(word)
        # pprint.pp(words)
        return words

    def unfinished_definitions(self) -> int:
        self.cursor.execute("SELECT word FROM dictionary WHERE definition IS NULL")
        unfinished_rows = self.cursor.fetchall()
        return len(unfinished_rows)

    @property
    def books(self):
        connection = sqlite3.connect(self.database_path)
        cursor = connection.cursor()
        cursor.execute("SELECT book_name FROM book_list")
        return_array = []
        for table in cursor.fetchall():
            return_array.append(table[0])
        return return_array

    @books.setter
    def books(self, value):
        self._books = value



import logging
import os

# TODO: I think that the entire book cleaner file is not needed.
#   Words should not be written back to a file. This is slow
CLEANED_FILES_DIRECTORY = "cleaned_files"
CLEANED_FILES_SUFFIX = "_cleaned"
# Sentences That should not be added to the completed file.
GARBAGE_SENTENCES = ["", "", ".", "。", "\n"]


class BookCleaner:
    def __init__(self, debug_level, force=False):
        """Internal tool used to sanatize input
        Use `clean(bookpath)` to sanatize files and remove junk data
        Args:
            `force`: Ignore confirmations on if you want to clean >50 files
        """
        self.logger = logging.getLogger("autoanki.bookcleaner")
        self.logger.setLevel(debug_level)
        self.file_list = []
        self.bookpath = ""
        self.force = force

    def clean(self, bookpath: str) -> None | list[str]:
        """
        Cleans the files contained in the bookpath. If bookpath is a single file, clean it.
        Args:
            `bookpath: filepath of files to be cleaned`
        Return:
            `str: list of cleaned file(s)`
        """
        if not os.path.exists(bookpath):
            self.logger.warning("Cannot find path [" + str(bookpath) + "]")
            return None
        self.logger.info("Bookpath: " + bookpath)

        # If the bookpath is a single file, clean and return it
        if os.path.isfile(bookpath):
            cleaned_files = [self._clean_file(bookpath, cleaned_files_root=None)]
            return cleaned_files
        else:
            dirty_files = []
            for root, dirs, files in os.walk(bookpath):
                for file in files:
                    # Only clean files that are not in cleaned_files directory
                    in_cleaned = str(root).find(CLEANED_FILES_DIRECTORY) != -1
                    if not in_cleaned:
                        if ".txt" in file:
                            dirty_files.append(os.path.join(root, file))

            # Check this cleaning won't be mean to the CPU
            if len(dirty_files) > 50 and not self.force:
                yn = input(
                    "Over 50 files. Are you sure you want to convert this many files? (Y/N)"
                )
                if yn.lower() != "y":
                    return None

        # Now we have a list of files to convert in a list
        cleaned_files = []

        # Create directory for files
        cleaned_files_root = os.path.join(bookpath, CLEANED_FILES_DIRECTORY)
        if not os.path.exists(cleaned_files_root):
            os.mkdir(cleaned_files_root)

        for file in dirty_files:
            cleaned_filepath = self._clean_file(
                file, cleaned_files_root=cleaned_files_root
            )
            if cleaned_filepath:
                cleaned_files.append(cleaned_filepath)

        return cleaned_files

    def _clean_file(self, filepath: str, cleaned_files_root: str):
        """
        Takes a txt file and cleans it up, filtering junk and putting every sentence on a new line
        Args:
            `filepath`: The txt file to clean
            `cleaned_files_root`: The root
        """

        # TODO: This will not work for books with nested files that share the same name
        filename = os.path.basename(filepath)
        new_filepath = os.path.join(cleaned_files_root, filename)
        self.logger.debug("Filepath:     " + filepath)
        self.logger.debug("New filepath: " + new_filepath)

        # Clean page file of characters that may cause issues.
        page_file = open(filepath, encoding="utf-8")
        try:
            page_sentences = page_file.read().split("。")
        except:
            self.logger.error(f"Critical error cleaning file: [{filepath}]")
            return
        cleaned_file = open(new_filepath, "w", encoding="utf-8")

        for page_sentence in page_sentences:
            # Clean string
            page_sentence = (
                page_sentence.lstrip()
                .rstrip()
                .replace("  ", " ")
                .lstrip('"')
                .rstrip('"')
                .replace("”", "'")
                .replace("　", "")
                .replace("“", "")
                .rstrip("。")
                .strip("'")
                .strip("'")
            )
            if page_sentence not in GARBAGE_SENTENCES:
                cleaned_file.write(page_sentence + "。" + "\n")

        return new_filepath



import os
import pytest

from autoanki.AutoAnki import AutoAnki

TEST_DB_NAME = "tests/UnitTest.db"


class TestAutoAnki:
    @classmethod
    def setUpClass(cls):  ### run once before all test cases ###
        pass

    @classmethod
    def tearDownClass(cls):  ### run once after all test cases ###
        pass

    def setUp(self):  ### run before each test case ###
        pass

    def tearDown(self):  ### run after each test case ###
        # If the test generated a database, delete it
        if os.path.exists(TEST_DB_NAME):
            os.remove(TEST_DB_NAME)  # one file at a time

    ### make sure to add => test_ <= as prefix to all test cases otherwise they won't work ###
    def test_init(self):
        aa = AutoAnki("zh", TEST_DB_NAME, debug_level=20)
        assert aa.book_list == []
        aa.print_database_info()

        self.tearDown()

    def test_book_list(self):
        aa = AutoAnki("zh", TEST_DB_NAME, debug_level=20)

        # No books
        assert len(aa.book_list) == 0

        # Add book
        aa.add_book_from_string("你好。再见", "Test Book")
        assert len(aa.book_list) == 1

        # Add book incorrectly
        aa.add_book_from_file("你好。再见", "Test Book")
        assert len(aa.book_list) == 1

        self.tearDown()



import unittest
from autoanki.AutoAnki import AutoAnki

# @unittest.skip('Some reason')


class TestDeckManager(unittest.TestCase):
    @classmethod
    def setUpClass(cls):  ### run once before all test cases ###
        pass

    @classmethod
    def tearDownClass(cls):  ### run once after all test cases ###
        pass

    def setUp(self):  ### run before each test case ###
        pass

    def tearDown(self):  ### run after each test case ###
        pass



import logging

from autoanki.BookCleaner import BookCleaner
from autoanki.DatabaseManager import ChineseDatabaseManager
from autoanki.DatabaseManager.DatabaseManager import DatabaseManager
from autoanki.Dictionary import CEDictionary
from autoanki.DeckManager import DeckManager
from autoanki.Dictionary.Dictionary import Dictionary
from autoanki.Tokenizer import ChineseTokenizer

import time

BLACK = "\u001b[30m"
RED = "\u001b[31m"
GREEN = "\u001b[32m"
YELLOW = "\u001b[33m"
BLUE = "\u001b[34m"
MAGENTA = "\u001b[35m"
CYAN = "\u001b[36m"
WHITE = "\u001b[37m"
RESET = "\u001b[0m"
logging.basicConfig(
    # filename='HISTORY.log',
    level=logging.WARNING,
    format=f"{GREEN}%(asctime)s{RESET} {RED}%(levelname)8s{RESET} {YELLOW}%(name)-16s{RESET}: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class LanguageResources:
    database_manager: DatabaseManager
    dictionary: Dictionary
    deck_manager: DeckManager


# class Chinese(LanguageResources):
# database_manager: ChineseDatabaseManager
# dictionary_manager: CEDictionary
# deck_manager: DeckManager

supported_languages = {"zh": "Yes"}


class AutoAnki:
    def __init__(
        self,
        language: str,
        database_filepath: str = "",
        debug_level=20,
        force=False,
        dictionary=None,
    ):
        """
        Creates an instance of autoanki.
        This creates a book cleaner, database connection, dictioary connection, and deck maker
        Args:
            `language`: The language to load, as per its ISO 639-1 two-letter code (en, zh, fr...)
            `database_filepath`: The filepath for the database. If none specified a new one will be created
            `logging_level`: between 0 (DEBUG) and 50(CRITICAL)
            `force`: Skip conformations for cleaning large numbers of files
        """
        self.logger = logging.getLogger("autoanki")
        self.logger.setLevel(debug_level)
        self.logger.debug(f"Autoanki logger active")

        if len(language) != 2:
            self.logger.warning(
                f"Incorrect language input: [{language}]. Please use it's 2-letter code"
            )
            return

        if not supported_languages.get(language):
            self.logger.warning(f"Unsupported language: [{language}]")
            return

        self.logger.info(
            f"===== {GREEN}Starting to load config: [{language}] {RESET}====="
        )
        total_start = time.time()
        self.force = force

        # self.book_cleaner = BookCleaner(debug_level, self.force)

        start = time.time()
        self.logger.info("Loading dictionary...")
        if dictionary:
            self.logger.info("Using custom dictionary")
        else:
            self.dictionary = CEDictionary(debug_level)
        end = time.time()
        self.logger.info(f"Done in {end - start:0.4f} seconds")

        self.database_filepath = database_filepath
        if database_filepath == "":
            self.logger.info("No database specified. Creating a new one...")
            timestr = time.strftime("%Y%m%d_%H%M%S")
            self.database_filepath = str("autoanki_" + timestr + ".db")
            ChineseDatabaseManager.create_database(self.database_filepath)
        else:
            if not ChineseDatabaseManager.is_database(self.database_filepath):
                self.logger.info("Creating database...")
                ChineseDatabaseManager.create_database(self.database_filepath)
                self.logger.info("Done creating database.")

        start = time.time()
        self.logger.info("Connecting to database...")
        self.database_manager = ChineseDatabaseManager(
            self.database_filepath, debug_level, dictionary=self.dictionary
        )
        end = time.time()
        self.logger.info(f"Done in {end - start:0.4f} seconds")

        start = time.time()
        self.logger.info("Connecting to DeckManager...")
        self.deck_manager = DeckManager(debug_level)
        end = time.time()
        self.logger.info(f"Done in {end - start:0.4f} seconds")

        total_end = time.time()
        self.logger.info(
            f"===== {GREEN}Done loading profile in {total_end - total_start:0.4f} seconds {RESET}====="
        )

    def add_book_from_string(self, contents: str, book_name: str = "Book Name"):
        """
        Add a directory full of files to the database
        Args:
            `contents`: path to the directory that contains the files to add
            `book_name`: The name of the book being added e.g. "Lost Prince"
        """
        self.logger.debug(f"autoanki: Adding book [{book_name}] from string")
        if not contents:
            self.logger.info(f"No contents supplied")
            return

        # Add the book to the database
        if not self.database_manager.add_book_from_string(contents, book_name):
            self.logger.warning("Unable to add [" + book_name + "] to database.")
            return

        self.logger.info("autoanki: Added book from string.")

    def add_book_from_file(self, filepath: str, book_name: str = "Book Name"):
        """
        Add a directory full of files to the database
        Args:
            `filepath`: path to the directory that contains the files to add
            `book_name`: The name of the book being added e.g. "Lost Prince"
        """
        self.logger.debug(
            f"autoanki: Adding book [{book_name}] from file: [{filepath}]"
        )
        if not filepath:
            self.logger.info(f"No filepath supplied")
            return

        # Add the book to the database
        if not self.database_manager.add_book_from_file(filepath, book_name):
            self.logger.warning("Unable to add [" + book_name + "] to database.")
            return

        self.logger.info("autoanki: Added [" + filepath + "].")

    def complete_unfinished_definitions(self):
        """
        autoanki contains an internal definitions table that is scraped from the internet. As words are added to
        autoanki, their definitions must be found.
        This function finds definitions and adds them to the table
        """
        self.logger.info("Checking for records...")
        self.database_manager.cursor.execute(
            "SELECT word FROM dictionary WHERE definition IS NULL"
        )
        response_rows = self.database_manager.cursor.fetchall()
        if len(response_rows) == 0:
            self.logger.info("No new rows to complete in dictionary table")
            return

        self.logger.info(
            "Adding " + str(len(response_rows)) + " rows to dictionary table"
        )
        self.tokenizer = ChineseTokenizer(dictionary=self.dictionary)
        for row in response_rows:
            word = str(row[0])

            # self.logger.debug(f"Finding: [{word}]")
            # self.logger.debug("Trying local dictionary...")
            params = self.dictionary.find_word(word)
            if params:
                # self.logger.debug(f"✅Found: [{params[8]}]")
                self.database_manager.update_definition(params)
                continue

            self.logger.info(f"❌Could not find: [{word}]")

    def deck_settings(
        self,
        inclue_traditional=True,
        inclue_part_of_speech=True,
        word_frequency_filter=None,
    ):
        """Configures settings for what's in the deck, and how it looks"

        `word_frequency_filter`: Float between 0 and 1. 1 being every word is included, 0 being none are included
        """
        self.deck_manager.settings(
            inclue_traditional,
            inclue_part_of_speech,
            word_frequency_filter,
        )

    def print_database_info(self):
        self.database_manager.print_info()

    @staticmethod
    def is_database(db_path):
        return ChineseDatabaseManager.is_database(db_path)

    @staticmethod
    def create_database(db_path: str):
        ChineseDatabaseManager.create_database(db_path)

    def create_deck(self, deck_name: str, filepath: str):
        """
        Creates a deck file in the directory of the main file.
        `deck_name` The name that will show up in Anki
        `filepath` Path to the file
        :return:
        """

        self.logger.info("Generating deck file [" + deck_name + ".apk]")
        words = self.database_manager.get_all_completed_definitions()

        deck_path = self.deck_manager.generate_deck_file(words, deck_name, filepath)
        if deck_path is None:
            self.logger.warning(
                "Was not able to create deck file for [", deck_name, "]"
            )
        else:
            self.logger.info("Generated deck file [" + deck_path + "]")

    @property
    def book_list(self):
        """
        Get a list of the books in the database
        :return: List of book names
        """
        return self.database_manager.books

    @book_list.setter
    def book_list(self, _):
        pass

    @property
    def unfinished_entries(self):
        return self.database_manager.unfinished_definitions()

    @unfinished_entries.setter
    def unfinished_entries(self, _):
        pass



from abc import ABC, abstractmethod
import unicodedata
import re
import os
import sqlite3


class DatabaseManager(ABC):
    @abstractmethod
    def __init__(self, debug_level=20):
        pass

    @staticmethod
    def convert_to_tablename(name: str) -> str:
        """Converts a string to a sql-valid table name"""
        value = unicodedata.normalize("NFKC", name)
        # value.replace("：",":")
        value = value.replace("：", "__")
        value = re.sub(r"[^\w\s-]", "", value.lower())
        return re.sub(r"[-\s]+", "_", value).strip("-_")

    @staticmethod
    def is_database(database_name: str):
        """Verifies integrity of AutoAnki database
        Args:
            `database_name`: Filepath of database
        """
        if not database_name.endswith(".db"):
            return False
        if not os.path.exists(database_name):
            return False
        try:
            connection = sqlite3.connect(database_name)
            cursor = connection.cursor()
            # This will fail if dictionary table does not exist
            cursor.execute("SELECT word FROM dictionary")
            connection.close()
        except sqlite3.OperationalError:
            return False
        return True

    @staticmethod
    def create_database(database_path: str) -> bool:
        pass

    @abstractmethod
    def print_info(self):
        pass

    @abstractmethod
    def add_contents_to_database(self, contents: str, table_name: str):
        pass

    @abstractmethod
    def add_book_from_string(self, contents: str, book_name: str) -> bool:
        pass

    @abstractmethod
    def add_book_from_file(self, filepath: str, book_name: str) -> bool:
        pass

    @abstractmethod
    def update_definition(self, params: list):
        pass

    @abstractmethod
    def get_all_completed_definitions(self) -> list:
        pass
